"""
Written by Matteo Dunnhofer - 2019

ResNet classes definitions
"""
import torch.nn as nn
from torchvision import models

class ResNet18(nn.Module):
    """
    ResNet18 definition
    """

    def __init__(self, cfg, training=False):
        super(ResNet18, self).__init__()

        self.name = 'ResNet18'
        self.is_training = training
        self.cfg = cfg

        resnet = models.resnet18(pretrained=True)
        modules = list(resnet.children())[:-2] # delete the last fc layer (and the avg pool for smaller input sizes)
        self.cnn_features = nn.Sequential(*modules)

        # fixed weights
        #for param in self.cnn_features.parameters():
        #    param.requires_grad = False

        del resnet



    def forward(self, x):
        """
        Builds the net

        Args:
            x: inputs to be predicted
        """
        return self.cnn_features(x).reshape(x.size(0), -1)

class ResNet18IF(nn.Module):
    """
    ResNet18 definition returning intermediate features
    """

    def __init__(self, cfg, training=False):
        super(ResNet18IF, self).__init__()

        self.name = 'ResNet18IF'
        self.is_training = training
        self.cfg = cfg

        resnet = models.resnet18(pretrained=True)
        modules = list(resnet.children())[:-2] # delete the last fc layer (and the avg pool for smaller input sizes)

        self.feat1 = nn.Sequential(*modules[:5])
        self.feat2 = nn.Sequential(*modules[5:6])
        self.feat3 = nn.Sequential(*modules[6:])

        # fixed weights
        """
        for param in self.feat1.parameters():
            param.requires_grad = False
        for param in self.feat2.parameters():
            param.requires_grad = False
        for param in self.feat3.parameters():
            param.requires_grad = False
        """

        del resnet



    def forward(self, x, linearize=False):
        """
        Builds the net

        Args:
            x: inputs to be predicted
        """
        f1 = self.feat1(x)
        f2 = self.feat2(f1)
        f3 = self.feat3(f2)

        if linearize:
            f2 = f2.view(f2.size(0), -1)
            f3 = f3.view(f3.size(0), -1)

        return f2, f3


class ResNet50(nn.Module):
    """
    ResNet50 definition
    """

    def __init__(self, cfg, training=False):
        super(ResNet50, self).__init__()

        self.name = 'ResNet50'
        self.is_training = training
        self.cfg = cfg

        resnet = models.resnet50(pretrained=True)
        modules = list(resnet.children())[:-2]  # delete the last fc layer (and the avg pool for smaller input sizes)

        self.cnn_features = nn.Sequential(*modules)

        # fixed weights
        # for param in self.cnn_features.parameters():
        #    param.requires_grad = False

        del resnet

    def forward(self, x):
        """
        Builds the net

        Args:
            x: inputs to be predicted
        """
        return self.cnn_features(x)
